"""Hypothesis strategies for mgf-http property tests (TS-13)."""

from __future__ import annotations

from hypothesis import strategies as st

# secret tokens unlikely to collide with scheme/host/path
secrets = st.text(alphabet="ABCDEFabcdef0123456789", min_size=8, max_size=32).map(
    lambda s: "SEKRET" + s
)
